
"""
GLOBAL LIST OF ALTERNATES AND SYNONYMS
"""

SHEETS = {
    "paleo": ["p", "paleo"],
    "chron": ["c", "chron"],
    "measurement": ["m", "meas", "measurement"],
    "ensemble": ["e", "ens", "ensemble"],
    "distribution": ["d", "dist", "distribution"],
    "summary": ["s", "sum", "summary"]
}

# TABLES
DATA_FRAMES = [
    "chron",
    "paleo",
    "ensemble",
    "calibratedage",
    "distribution",
    "model",
    "table",
    "measurement"
]

# Allowable int and float fields
NUMERIC_INTS = [
    "maxYear",
    "minYear",
    "pubYear",
    "number",
    "year",
    "age14c",
    "sd14c"
]

NUMERIC_FLOATS = [
    "LiPDVersion",
    "coordinates",
    "depth"
]

# FILTER
COMPARISONS = {
    "==": "=",
    "is": "=",
    "greater than": ">",
    "equals": "=",
    "equal": "=",
    "less than": "<"
}

# EXCEL
EXCEL_GEO = [
    "latMax",
    "latMin",
    "lonMax",
    "lonMin",
    "elevation",
    "siteName"
]


# Missing value name appears as many variations. Try to account for all of them
ALTS_MV = [
    'missing value',
    'missing values',
    'missingvalue',
    'missingvalues',
    'missing_values',
    'missing variables',
    'missing_variables',
    'missingvariables'
]

# NOAA
# Lists for what keys go in specific dictionary blocks
NOAA_SITE_INFO = {
    'lat': [
        'northernmostlatitude',
        'northernmost latitude',
        'northernmost_latitude',
        'southernmostlatitude',
        'southernmost latitude',
        'southernmost_latitude'],
    'lon': [
        'easternmostlongitude',
        'easternmost longitude',
        'easternmost_longitude',
        'westernmostlongitude',
        'westernmost longitude',
        'westernmost_longitude'],
    'properties': [
        'location',
        'country',
        'elevation',
        'site_name',
        'region'
    ],
}

FUNDING_LIST = [
    'funding_agency_name',
    'grant'
]

# 13 is a list of keys to ignore when using create_blanks
NOAA_SECTIONS = {
    1: ['onlineResource', 'studyName', 'archive', 'parameterKeywords', 'originalSourceUrl'],
    2: ['date'],
    3: ['studyName'],
    4: ['investigators'],
    5: ['description'],
    6: ['pub'],
    7: ['funding', 'agency', 'grant'],
    8: ['geo'],
    9: ['collectionName', 'earliestYear', 'mostRecentYear', 'timeUnit', 'coreLength', 'notes'],
    10: ['speciesName', 'commonName', 'treeSpeciesCode'],
    11: ['chronData'],
    12: ['paleoData'],
    13: []
}

# The order of the items in the list is the order that we want to write them to the file.
# 11 is the order for writing each column in the variables section
NOAA_ORDERING = {
    1: ['studyName',
        'onlineResource',
        'originalSourceUrl',
        'archive',
        'parameterKeywords'],
    2: ['date'],
    3: ['studyName'],
    4: ['investigators'],
    5: ['description'],
    6: ['author',
        'publishedDateOrYear',
        'publishedTitle',
        'journalName',
        'volume',
        'edition',
        'issue',
        'pages',
        "reportNumber",
        'doi',
        'onlineResource',
        'fullCitation',
        'abstract'
        ],
    7: ['agency', 'grant'],
    8: ['siteName',
        'location',
        'country',
        'northernmostLatitude',
        'southernmostLatitude',
        'easternmostLongitude',
        'westernmostLongitude',
        'elevation'],
    9: ['collectionName',
        'earliestYear',
        'mostRecentYear',
        'timeUnit',
        'coreLength',
        'notes'],
    10: ['speciesName', 'commonName'],
    11: ["variableName",
         "variableType",
         "measurementMaterial",
         "uncertainty",
         "units",
         "seasonality",
         "archive",
         "detail",
         "measurementMethod",
         "dataType"
         ]
}

UNITS = {
    "meters": "m",
    "meter": "m",
    "centimeters": "cm",
    "centimeter": "cm",
    "feet": "ft",
    "foot": "ft",
    "millimeter": "mm",
    "millimeters": "mm"
}

# These keys do not have a corresponding match in the LiPD schema. Do not include them.
NOAA_IGNORE_KEYS = [
    "originalSourceUrl",
    "original_source_url",
    "commonName",
    "common_name",
    "parameterKeywords",
    "parameter_keywords",
    "speciesName",
    "species_name",
    "timeUnit",
    "time_unit",
    "collectionName",
    "collection_name",
    "earliestYear",
    "earliest_year",
    "mostRecentYear",
    "most_recent_year",
    "coreLength",
    "core_length",
    "note"
]

NOAA_DOI_KEYS = [
    "DOI",
    "DOI_url",
    "DOI_id"
]

# Excel metadata keys. Left side: Excel keys, Right Side: LiPD keys
EXCEL_KEYS = {
    "archive type": "archiveType",
    "dataset name": "dataSetName",
    "metadata": "metadata",
    "chronology": "chronology",
    "data (qc)": "dataQC",
    "data(qc)": "dataQC",
    "data (original)": "dataOriginal",
    "data(original)": "dataOriginal",
    "data": "data",
    "proxyList": "proxy",
    "proxy": "proxy",
    "about": "about",
    "study title": "studyName",
    "investigators": "investigators",
    "authors": "author",
    "publication title": "title",
    "journal": "journal",
    "year": "year",
    "volume": "volume",
    "issue": "issue",
    "pages": "pages",
    "dataUrl": "dataUrl",
    "link": "link",
    "report number": "reportNumber",
    "doi": "id",
    "abstract": "abstract",
    "alternate citation": "alternateCitation",
    "site name": "siteName",
    "northernmost latitude": "latMax",
    "southernmost latitude": "latMin",
    "easternmost longitude": "lonMax",
    "westernmost longitude": "lonMin",
    "elevation": "elevation",
    "variablename": "variableName",
    "description": "description",
    "short_name": "variableName",
    "what": "description",
    "material": "measurementMaterial",
    "error": "error",
    "units": "units",
    "seasonality": "seasonality",
    "archive": "archive",
    "detail": "detail",
    "method": "method",
    "data_type": "dataType",
    "basis of climate relation": "basis",
    "climate_interpretation_code": "climateInterpretation",
    "climate_intepretation_code": "climateInterpretation",
    "notes": "notes",
    "comments": "notes"
}

EXCEL_TEMPLATE = [
    "data",
    "chronology table containing measured depths and ages",
    "chronology table as paragraph (insert table into cell b3) not preferred",
    "chronology table (use as many rows and columns are needed below this line)",
    "note: additional chronology tables should be put below the first one with no blank lines.",
    "variables",
    "use one row to define each variable, beginning with depth then age; add additional worksheets for additional tables",
    "note: data_type is 'n' for numeric and 'c' for character data",
    "paste data table below starting in column a",
    "the value or character string used as a placeholder for missing values",
    "<missing value>",
    "<notes that describe the table as a whole>"
]

EXCEL_HEADER = [
    "variablename",
    "short_name",
    "what",
    "material",
    "error",
    "units",
    "seasonality",
    "archive",
    "detail",
    "method",
    "data_type",
    "uncertainty_level",
    "calibration_curve",
    "climate_interpretation_code",
    "standard",
    "basis_of_climate_relation"
]


import csv

from ..helpers.loggers import *
from ..helpers.misc import cast_value

logger_csvs = create_logger("csvs")


# Merge CSV and Metadata


def merge_csv_metadata(d):
    """
    Using the given metadata dictionary, retrieve CSV data from CSV files, and insert the CSV
    values into their respective metadata columns. Checks for both paleoData and chronData tables.
    :param dict d: Metadata
    :return dict: Modified metadata dictionary
    """
    logger_csvs.info("enter import_csv_to_metadata")

    # Add CSV to paleoData
    if "paleoData" in d:
        d["paleoData"] = _merge_csv_section(d["paleoData"], "paleo")

    # Add CSV to chronData
    if "chronData" in d:
        d["chronData"] = _merge_csv_section(d["chronData"], "chron")

    logger_csvs.info("exit import_csv_to_metadata")
    return d


def _merge_csv_section(section_data, pc):
    """
    Add csv data to all paleo data tables
    :param dict section_data: Metadata
    :return dict: Modified metadata
    """
    logger_csvs.info("enter import_csv: {}".format(pc))

    try:
        # Loop through each table_data in paleoData
        for table_name, table_data in section_data.items():

            if "{}MeasurementTable".format(pc) in table_data:
                # Send whole table_data through. Adds csv data to columns
                for table_name2, table_data2 in table_data["{}MeasurementTable".format(pc)].items():
                    table_data["{}MeasurementTable".format(pc)][table_name2] = _add_csv_to_columns(table_data2)

            if "{}Model".format(pc) in table_data:
                table_data["{}Model".format(pc)] = _merge_csv_model(table_data["{}Model".format(pc)], pc)
    except AttributeError:
        print("Error: {} section must be a dictionary data type".format(pc))

    logger_csvs.info("exit import_csv: {}".format(pc))
    return section_data


def _merge_csv_model(models, pc):
    """
    Add csv data to each column in chron model
    :param list models: Metadata
    :param str pc: Paleo or chron
    :return dict: Modified metadata
    """
    logger_csvs.info("enter import_model_csv: {}".format(pc))

    try:
        for model in models:

            if "{}ModelTable".format(pc) in model:
                model["{}ModelTable".format(pc)] = _add_csv_to_columns(model["{}ModelTable".format(pc)])
            if "ensembleTable" in model:
                model["ensembleTable"] = _add_csv_to_columns(model["ensembleTable"])

            # Check for calibratedAges (old format) and distributionTable (current format)
            if "calibratedAges" in model:
                model["distributionTable"] = {}
                # Calibrated age tables are nested. Go down an extra layer.
                for k, v in model["calibratedAges"].items():
                    model["distributionTable"][k] = _add_csv_to_columns(v)
            elif "distributionTable" in model:
                for k, v in model["distributionTable"].items():
                    model["distributionTable"][k] = _add_csv_to_columns(v)
    except AttributeError:
        print("Error: Model section must be a list data type")

    logger_csvs.info("exit import_model_csv: {}".format(pc))
    return models


def _add_csv_to_columns(table):
    """
    Add csv data to each column in a list of columns
    :param dict table: Table metadata
    :return dict: Table metadata with csv "values" entry
    """
    # Get the filename or this table
    filename = _merge_get_filename(table)

    # If there's no filename, bypass whole process
    if filename:
        # Call read_csv_to_columns for this filename. csv_data is list of lists.
        csv_data = _read_csv_from_file(filename)

        # Start putting CSV data into corresponding column "values" key
        try:
            for col_name, col in table['columns'].items():
                # The "number" entry in the ensemble table is a list of columns, instead of an int.
                # In this case, just store all the csv data as a batch.
                if isinstance(col["number"], list):
                    col["values"] = csv_data
                # For all other cases, "number" is a single int, and "values" should hold one column list.
                else:
                    col['values'] = csv_data[col["number"] - 1]
        except IndexError:
            logger_csvs.warning("add_csv_to_json: IndexError: index out of range of csv_data list")
        except KeyError:
            logger_csvs.debug("add_csv_to_json: KeyError: missing columns key")

    return table


def _read_csv_from_file(filename):
    """
    Opens the target CSV file and creates a dictionary with one list for each CSV column.
    :param str filename:
    :return list of lists: column values
    """
    logger_csvs.info("enter read_csv_to_columns")
    d = {}
    l = []
    try:
        with open(filename, 'r') as f:
            r = csv.reader(f, delimiter=',')

            # Create a dict with X lists corresponding to X columns
            for idx, col in enumerate(next(r)):
                d[idx] = []
                d[idx] = cast_value(col)

            # Start iter through CSV data
            for row in r:
                for idx, col in enumerate(row):
                    # Append the cell to the correct column list
                    d[idx] = cast_value(col)

        # Make a list of lists out of the dictionary instead
        for idx, col in d.items():
            l.append(col)

    except FileNotFoundError as e:
        print('CSV FileNotFound: ' + filename)
        logger_csvs.warn("read_csv_to_columns: FileNotFound: {}, {}".format(filename, e))
    logger_csvs.info("exit read_csv_to_columns")
    return l


# SKIM CSV


def get_csv_from_metadata(metadata):
    """
    Get only csv data from metadata
    :param dict metadata: Metdata
    :return dict: Csv Data
    """
    logger_csvs.info("enter get_organized_csv")

    d = {}
    crumbs = _get_datasetname(metadata)

    if "paleoData" in metadata:
        crumbs_tmp = "{}paleo".format(crumbs)
        pd_out = _get_csv_section(metadata["paleoData"], crumbs_tmp, "paleo")
        d.update(pd_out)

    if "chronData" in metadata:
        crumbs_tmp = "{}chron".format(crumbs)
        cd_out = _get_csv_section(metadata["chronData"], crumbs_tmp, "chron")
        d.update(cd_out)

    logger_csvs.info("exit get_organized_csv")
    return d


def _get_csv_section(section_data, crumbs, pc):
    """
    Get table name, variable name, and column values from paleo metadata
    :param dict section_data: Metadata
    :param str crumbs: Table name crumbs
    :param str pc: Paleo or chron
    :return dict: Metadata
    """
    logger_csvs.info("enter get_csv_section: {}".format(pc))
    d = {}
    try:
        # Process the tables in chronData
        for name_table, data_table in section_data.items():
            crumbs_tmp = "{}{}".format(crumbs, str(name_table))

            # Process each entry sub-table below if they exist
            if "{}MeasurementTable".format(pc) in data_table:
                idx = 1
                for name_pmt, data_pmt in data_table["{}MeasurementTable".format(pc)].items():
                    crumbs_tmp_cmt = "{}{}{}csv".format(crumbs_tmp, "{}MeasurementTable".format(pc), idx)
                    filename = _export_get_filename(data_pmt, crumbs_tmp_cmt)
                    out = _search_table_for_vals(data_pmt)
                    d[filename] = out
                    idx += 1

            if "{}Model".format(pc) in data_table:
                for item in data_table["{}Model".format(pc)]:
                    if "calibratedAges" in item:
                        # CA has an extra level of nesting
                        for name_ca, data_ca in item["calibratedAges"].items():
                            crumbs_tmp_ca = "{}{}.csv".format(crumbs, name_ca)
                            filename = _export_get_filename(data_ca, crumbs_tmp_ca)
                            out = _search_table_for_vals(data_ca)
                            d[filename] = out
                    elif "distributionTable" in item:
                        # CA has an extra level of nesting
                        for name_ca, data_ca in item["distributionTable"].items():
                            crumbs_tmp_ca = "{}{}.csv".format(crumbs, name_ca)
                            filename = _export_get_filename(data_ca, crumbs_tmp_ca)
                            out = _search_table_for_vals(data_ca)
                            d[filename] = out

                    if "{}ModelTable".format(pc) in item:
                        crumbs_tmp_cmt2 = "{}{}.csv".format(crumbs, "{}ModelTable".format(pc))
                        filename = _export_get_filename(item["{}ModelTable".format(pc)], crumbs_tmp_cmt2)
                        out = _search_table_for_vals(item["{}ModelTable".format(pc)])
                        d[filename] = out

                    elif "summaryTable".format() in item:
                        crumbs_tmp_cmt2 = "{}{}.csv".format(crumbs, "summaryTable".format(pc))
                        filename = _export_get_filename(item["summaryTable"], crumbs_tmp_cmt2)
                        out = _search_table_for_vals(item["summaryTable"])
                        d[filename] = out

                    if "ensembleTable" in item:
                        crumbs_tmp_et = "{}{}.csv".format(crumbs, "ensembleTable")
                        filename = _export_get_filename(item["ensembleTable"], crumbs_tmp_et)
                        out = _search_table_for_vals(item["ensembleTable"])
                        d[filename] = out

    except AttributeError:
        logger_csvs.info("get_csv_section: {}, AttributeError: expected type dict, given type {}".format(pc, type(section_data)))

    logger_csvs.info("exit get_section_csv: {}".format(pc))
    return d


def write_csv_to_file(d):
    """
    Writes columns of data to a target CSV file.
    :param dict d: A dictionary containing one list for every data column. Keys: int, Values: list
    :return None:
    """
    logger_csvs.info("enter write_csv_to_file")
    for filename, data in d.items():
        l_columns = []
        for k, v in data.items():
            l_columns.append(v)
        rows = zip(*l_columns)
        with open(filename, 'w+') as f:
            w = csv.writer(f)
            for row in rows:
                w.writerow(row)
    logger_csvs.info("exit write_csv_to_file")
    return


# HELPERS


def _search_table_for_vals(d):
    """
    Search a data tables for column values. Return a dict of column values
    :param dict d: Table data
    :return dict: Column values. ref by var name
    """
    cols = {}
    if "columns" in d:
        try:
            for name_col, data_col in d["columns"].items():
                vals = _search_col_for_vals(data_col)
                if vals:
                    cols[name_col] = vals
        except AttributeError:
            print("Error: Table 'columns' entries must be a dictionary type")
            logger_csvs.debug("search_table_for_vals: AttributeError: expected type dict, given type {}".format(type(d)))

    return cols


def _search_col_for_vals(data_col):
    """
    Get the values key from a data column
    :param dict data_col:
    :return list:
    """
    val = []
    try:
        val = data_col["values"]
    except KeyError:
        print("Error: Data column is missing a 'values' entry")
        logger_csvs.info("get_values_key: KeyError: missing values key")

    return val


def _merge_get_filename(table):
    """
    Get the value from the filename field if the table has one
    :param dict table: Table metadata
    :return str: Filename
    """
    file = ""
    try:
        file = table["filename"]
    except KeyError:
        print("Error: Data table is missing a csv filename")
        logger_csvs.info("get_table_csv_filename: KeyError: missing csv filename")
    return file


def _export_get_filename(table, crumbs):
    """
    Get the filename from a data table. If it doesn't exist, create a new one based on table hierarchy in metadata file.
    format: <dataSetName>.<dataTableType><idx>.<dataTableName>.csv
    example: ODP1098B.Chron1.ChronMeasurementTable.csv
    :param dict table: Table data
    :return str: Filename
    """
    try:
        filename = table["filename"]
    except KeyError:
        logger_csvs.info("get_filename: KeyError: missing filename key")
        filename = crumbs
    return filename


def _get_datasetname(d):
    """
    Get data set name from metadata
    :param dict d: Metadata
    :return str: Data set name
    """
    try:
        s = d["dataSetName"]
    except KeyError:
        logger_csvs.warn("get_data_set_name: KeyError: missing dataSetName")
        s = "lipd"
    return s


def _merge_ensemble(ensemble, col_nums, col_vals):
    """
    The second column is not typical.
    "number" is a list of column numbers and "values" is an array of column values.
    Before we can write this to csv, it has to match the format the writer expects.
    :param dict ensemble: First column data
    :param list col_nums: Second column numbers. list of ints
    :param list col_vals: Second column values. list of lists
    :return dict:
    """
    try:
        # Loop for each column available
        for num in col_nums:
            # first column number in col_nums is usually "2", so backtrack once since ensemble already has one entry
            # col_vals is 0-indexed, so backtrack 2 entries
            ensemble[num-1] = col_vals[num - 2]

    except IndexError:
        logger_csvs.debug("merge_ensemble: IndexError: index out of range")

    return ensemble


# EXPORT


# def get_csv_from_metadata(d):
#     """
#     Initial call. Start the process of finding column values in metadata and writing it to csv files
#     :param dict d:
#     :return:
#     """
#     logger_csvs.info("enter export_csv_to_metadata")
#     csv_only = {}
#     crumbs = _get_datasetname(d)
#
#     # read from paleo data
#     if "paleoData" in d:
#         tmp = _get_csv_section(d["paleoData"], crumbs, "paleo")
#         csv_only.update(tmp)
#
#     # read from chron data
#     if "chronData" in d:
#         tmp = _get_csv_section(d["chronData"], crumbs, "chron")
#         csv_only.update(tmp)
#
#     # Write out all the tables that we collected
#     for filename, data in csv_only.items():
#         write_csv_to_file(filename, data)
#
#     logger_csvs.info("exit export_csv_to_metadata")
#     return csv_only
#
#
# def _get_csv_section(section_data, crumbs, pc):
#     """
#     Export data from chron data
#     :param dict section_data: Metadata
#     :param str crumbs: Table name bread crumbs
#     :param str pc: Paleo or chron
#     :return:
#     """
#     logger_csvs.info("enter export_csv_section: {}".format(pc))
#     chron_csv = {}
#     idx = 1
#
#     try:
#         for table_name, table_data in section_data.items():
#             # Process Model
#             if "{}Model".format(pc) in table_data:
#
#                 # String together crumbs for this level
#                 crumbs_tmp = "{}{}{}".format(crumbs, pc, str(idx))
#
#                 # Replace the chronModel in-place
#                 tmp = _get_csv_model(table_data["{}Model".format(pc)], crumbs_tmp, pc)
#                 chron_csv.update(tmp)
#
#             # Process chronMeasurementTable
#             if "{}MeasurementTable".format(pc) in table_data:
#
#                 # String together crumbs for this level
#                 crumbs_tmp = "{}{}{}{}MeasurementTable".format(crumbs, pc, str(idx), pc)
#
#                 # Replace the chronMeasurementTable in-place
#                 tmp = _get_csv_from_cols(table_data["{}MeasurementTable".format(pc)], crumbs_tmp)
#                 chron_csv.update(tmp)
#
#             idx += 1
#
#     except AttributeError:
#         print("Error: {} section must be a dictionary data type".format(pc))
#         logger_csvs.debug("export_csv_section: {}, AttributeError: expected dict type, given {}".format(pc, type(section_data)))
#
#     logger_csvs.info("exit export_csv_section: {}".format(pc))
#     return chron_csv
#
#
# def _get_csv_model(models, crumbs, pc):
#     """
#     Export data from chron model
#     :param dict models: Metadata
#     :param str crumbs: Table name bread crumbs
#     :param str pc: Paleo or chron
#     :return dict:
#     """
#     logger_csvs.info("enter export_model_csv: {}".format(pc))
#     model_csv = {}
#     idx = 1
#
#     for model in models:
#         tmp = {}
#         dist = ""
#
#         # String together crumbs for this level
#         crumbs_tmp = "{}model{}".format(crumbs, str(idx))
#
#         # This could be more efficient by doing model_csv.update(_rip_csv_to_columns(model["chronModelTable"]))
#         if "{}ModelTable" in model:
#             crumbs_model = "{}{}summaryTable.csv".format(crumbs_tmp, pc)
#             tmp = _get_csv_from_cols(model["{}ModelTable".format(pc)], crumbs_model)
#         elif "summaryTable" in model:
#             crumbs_model = "{}{}summaryTable.csv".format(crumbs_tmp, pc)
#             tmp = _get_csv_from_cols(model["summaryTable".format(pc)], crumbs_model)
#         if "ensembleTable" in model:
#             crumbs_ensemble = "{}ensembleTable.csv"
#             tmp = _get_csv_from_cols(model["ensembleTable"], crumbs_ensemble)
#         if "calibratedAges" in model:
#             dist = "calibratedAges"
#         elif "distributionTable" in model:
#             dist = "distributionTable"
#         if dist:
#             # Calibrated age tables are nested. Go down an extra layer.
#             c_idx = 1
#             for k, v in model[dist].items():
#                 crumbs_ca = "{}distributionTable{}.csv".format(crumbs_tmp, str(c_idx))
#                 tmp2 = _get_csv_from_cols(v, crumbs_ca)
#                 tmp.update(tmp2)
#                 c_idx += 1
#
#         # Update the model_csv with the new csv data that trickled up.
#         model_csv.update(tmp)
#         idx += 1
#
#     logger_csvs.info("exit export_model_csv: {}".format(pc))
#     return model_csv
#
#
# def _get_csv_from_cols(table, crumbs):
#     """
#     Rip the csv data from a given table
#     :param dict table: Table data
#     :return dict: key: table filename val: column data
#     """
#     col_data = {}
#     complete = {}
#
#     # Get the filename or table name
#     filename = _export_get_filename(table, crumbs)
#
#     # If columns is here, start parsing.
#     if "columns" in table:
#         col_data = _get_csv_from_cols_sub(table)
#
#     # If this file had a table name, then there's an extra level before columns.
#     else:
#         try:
#             for t_name, t_data in table.items():
#                 if "columns" in t_data:
#                     col_data = _get_csv_from_cols_sub(t_data)
#         except KeyError:
#             logger_csvs.warning("rip_csv_from_columns: KeyError: missing number or values keys")
#         except AttributeError:
#             logger_csvs.warning("rip_csv_from_columns: AttributeError: expected dictionary type, given {}".format(type(table)))
#
#     complete[filename] = col_data
#     return complete
#
#
# def _get_csv_from_cols_sub(table):
#     """
#     Sub routine to parse column data.
#     :param dict table: Metadata
#     :return dict: Metadata
#     """
#     col_data = {}
#     keys = {"number": None, "values": None}
#     ens = False
#     if "columns" in table:
#         # Loop over "columns" key and track all data
#         for var, col in table["columns"].items():
#
#             for k, v in keys.items():
#                 try:
#                     keys[k] = col[k]
#                 except KeyError:
#                     print("Error: missing {} key in column".format(k))
#                     logger_csvs.warning("rip_csv_to_columns: KeyError: missing {} key".format(k))
#
#             if keys["number"] and keys["values"]:
#                 # Get the column number, and decrement for 0-index.
#                 # This is the position that the col will be written to in csv.
#                 # Catch the ensemble table exception. "number" is a list instead of an int
#                 vals = keys["values"]
#                 num = keys["number"]
#
#                 if isinstance(keys["number"], list):
#                     # Set ensemble flag so we know to process data further
#                     ens = True
#                 else:
#                     # Set data to the running col data. No further processing needed
#                     col_data[num-1] = vals
#
#         if ens:
#             col_data = _merge_ensemble(col_data, num, vals)
#
#     return col_data
#
#



import csv

import xlrd

from ..doi.doi_resolver import *
from ..helpers.bag import *
from ..helpers.directory import *
from ..helpers.zips import *
from ..helpers.blanks import *
from ..helpers.loggers import *
from ..helpers.alternates import *
from ..helpers.regexes import RE_SHEET


logger_excel = create_logger('excel_main')


def excel():
    """
    Parse data from Excel spreadsheets into LiPD files.
    :return:
    """
    dir_root = os.getcwd()

    # Find excel files
    f_list = list_files('.xls') + list_files('.xlsx')
    print("Found " + str(len(f_list)) + " Excel files")
    logger_excel.info("found excel files: {}".format(len(f_list)))

    # Run once for each file
    for name_ext in f_list:

        # Filename without extension
        name = os.path.splitext(name_ext)[0]
        name_lpd = name + '.lpd'
        print("processing: {}".format(name_ext))
        logger_excel.info("processing: {}".format(name_ext))

        # Create a temporary folder and set paths
        dir_tmp = create_tmp_dir()
        dir_bag = os.path.join(dir_tmp, name)
        dir_data = os.path.join(dir_bag, 'data')

        # Make folders in tmp
        os.mkdir(os.path.join(dir_bag))
        os.mkdir(os.path.join(dir_data))

        pending_csv = []
        paleo_ct = 1
        chron_ct = 1
        sheets = []
        final = OrderedDict()
        logger_excel.info("variables initialized")

        """
        EACH DATA TABLE WILL BE STRUCTURED LIKE THIS
        "paleo_chron": "paleo",
        "pc_idx": 1,
        "model_idx": "",
        "table_type": "measurement",
        "table_idx": 1,
        "name": sheet,
        "filename": sheet,
        "data": { column data }
        """

        # Open excel workbook with filename
        try:
            workbook = xlrd.open_workbook(name_ext)
            logger_excel.info("opened XLRD workbook")

        except Exception as e:
            # There was a problem opening a file with XLRD
            print("Failed to open Workbook: {}".format(name))
            logger_excel.debug("excel: xlrd failed to open workbook: {}, {}".format(name, e))

        if workbook:
            # Check what worksheets are available, so we know how to proceed.
            for sheet in workbook.sheet_names():

                # Use this for when we are dealing with older naming styles of "data (qc), data, and chronology"
                old = "".join(sheet.lower().strip().split())

                # Don't parse example sheets. For user use only.
                if "example" not in sheet and "sample" not in sheet:
                    # Group the related sheets together, so it's easier to place in the metadata later.
                    if 'metadata' in sheet.lower():
                        metadata_str = sheet
                    elif "about" not in sheet.lower() and "proxy" not in sheet.lower():
                        logger_excel.info("creating sheets metadata")
                        m = re.match(RE_SHEET, sheet.lower())
                        if m:
                            sheets, paleo_ct, chron_ct = _build_sheet(m, sheets, sheet, name, paleo_ct, chron_ct)

                        elif old == "data" or old in "data(qc)":
                            sheets.append({
                                "paleo_chron": "paleo",
                                "pc_idx": paleo_ct,
                                "model_idx": None,
                                "table_type": "measurement",
                                "table_idx": 1,
                                "name": sheet,
                                "filename": "{}paleo{}measurementTable1.csv".format(name, paleo_ct),
                                "table_name": "paleo{}measurementTable1".format(paleo_ct),
                                "data": ""
                            })
                            paleo_ct += 1
                        elif old == "chronology":
                            sheets.append({
                                "paleo_chron": "chron",
                                "pc_idx": chron_ct,
                                "model_idx": None,
                                "table_type": "measurement",
                                "table_idx": 1,
                                "name": sheet,
                                "filename": "{}chron{}measurementTable1.csv".format(name, chron_ct),
                                "table_name": "chron{}measurementTable1".format(chron_ct),
                                "data": ""
                            })
                            chron_ct += 1
                        else:
                            # Sheet name does not conform to standard. Guide user to create a standardized sheet name.
                            print("This sheet name does not conform to naming standard: {}".format(sheet))
                            sheets, paleo_ct, chron_ct = _prompt_sheets(sheets, sheet, name, paleo_ct, chron_ct)

            # METADATA WORKSHEETS
            # Parse Metadata sheet and add to output dictionary
            if metadata_str:
                logger_excel.info("parsing worksheet: {}".format(metadata_str))
                cells_dn_meta(workbook, metadata_str, 0, 0, final)

            # PALEO AND CHRON SHEETS
            for sheet in sheets:
                logger_excel.info("parsing data worksheet: {}".format(sheet))
                sheet_meta, sheet_csv = _parse_sheet(name, workbook, sheet)
                if sheet_csv and sheet_meta:
                    pending_csv.append(sheet_csv)
                    sheet["data"] = sheet_meta

            # Reorganize sheet metadata into LiPD structure
            d_paleo, d_chron = _place_tables(sheets)

            # Add organized metadata into final dictionary
            final['paleoData'] = d_paleo
            final['chronData'] = d_chron

            # OUTPUT

            # Create new files and dump data in dir_data
            os.chdir(dir_data)

            # WRITE CSV
            _write_data_csv(pending_csv)

            # JSON-LD
            # Invoke DOI Resolver Class to update publisher data
            try:
                logger_excel.info("invoking doi resolver")
                final = DOIResolver(dir_root, name, final).main()
            except Exception as e:
                print("Error: doi resolver failed: {}".format(name))
                logger_excel.debug("excel: doi resolver failed: {}, {}".format(name, e))

            # Dump final_dict to a json file.
            final["LiPDVersion"] = 1.2
            write_json_to_file(name + '.jsonld', final)

            # Move files to bag root for re-bagging
            # dir : dir_data -> dir_bag
            logger_excel.info("start cleanup")
            dir_cleanup(dir_bag, dir_data)

            # Create a bag for the 3 files
            finish_bag(dir_bag)

            # dir: dir_tmp -> dir_root
            os.chdir(dir_root)

            # Check if same lpd file exists. If so, delete so new one can be made
            if os.path.isfile(name_lpd):
                os.remove(name_lpd)

            # Zip dir_bag. Creates in dir_root directory
            logger_excel.info("re-zip and rename")
            zipper(dir_tmp, name, name_lpd)
            os.rename(name_lpd + '.zip', name_lpd)

        # Move back to dir_root for next loop.
        os.chdir(dir_root)

        # Cleanup and remove tmp directory
        shutil.rmtree(dir_tmp)

    print("Process Complete")
    return


# SORT PARSED DATA

def _prompt_sheets(sheets, old_sheet, name, paleo_ct, chron_ct):
    """
    Guide the user to create a proper, standardized sheet name
    :param list sheets: Running list of sheet metadata
    :param str old_sheet: Original sheet name
    :param str name: Data set name
    :param int paleo_ct: Running count of paleoData tables
    :param int chron_ct: Running count of chronData tables
    :return sheets paleo_ct chron_ct: Updated sheets and counts
    """
    cont = True
    # Loop until valid sheet name is built, or user gives up
    while cont:
        pc = input("Is this a (p)aleo or (c)hronology sheet?").lower()
        if pc in ("p", "c", "paleo", "chron", "chronology"):
            tt = input("Is this a (d)istribution, (e)nsemble, (m)easurement, or (s)ummary sheet?").lower()
            if tt in SHEETS["distribution"] or tt in SHEETS["ensemble"] \
                    or tt in SHEETS["summary"] or tt in SHEETS["measurement"]:
                # valid answer, keep going
                if tt in SHEETS["distribution"]:
                    tt = "distribution"
                elif tt in SHEETS["summary"]:
                    tt = "summary"
                elif tt in SHEETS["ensemble"]:
                    tt = "ensemble"
                elif tt in SHEETS["measurement"]:
                    tt = "measurement"
                if pc in SHEETS["paleo"]:
                    sheet = "{}{}{}{}".format("paleo", paleo_ct, tt, 1)
                elif pc in SHEETS["chron"]:
                    sheet = "{}{}{}{}".format("chron", chron_ct, tt, 1)

                # Test the sheet that was built from the user responses.
                # If it matches the Regex, then continue to build the sheet metadata. If not, try again or skip sheet.
                m = re.match(RE_SHEET, sheet.lower())
                if m:
                    sheets, paleo_ct, chron_ct = _build_sheet(m, sheets, old_sheet, name, paleo_ct, chron_ct)
                    print("Sheet created: {}".format(sheet))
                    cont = False
                else:
                    resp = input("invalid sheet name. try again? (y/n): ")
                    if resp == "n":
                        print("No valid sheet name was created. Skipping sheet: {}".format(sheet))
                        cont = False

    print("=====================================================")
    return sheets, paleo_ct, chron_ct


def _build_sheet(m, sheets, old_sheet, name, paleo_ct, chron_ct):
    """
    Build metadata for a sheet. Receive valid regex match object and use that to create metadata.
    :param obj m: Regex match object
    :param list sheets: Running list of sheet metadata
    :param str old_sheet: Original sheet name
    :param str name: Data set name
    :param int paleo_ct: Running count of paleoData tables
    :param int chron_ct: Running count of chronData tables
    :return sheets paleo_ct chron_ct: Updated sheets and counts
    """
    model_idx = None
    pc = m.group(1)
    # Get the model idx number from string if it exists
    if m.group(3):
        model_idx = int(filter(str.isdigit, m.group(3)))
    if pc == "paleodata" or pc == "paleo":
        pc = "paleo"
        paleo_ct += 1
    elif pc == "chrondata" or pc == "chron":
        pc = "chron"
        chron_ct += 1

    # Standard naming. This matches the regex and the sheet name is how we want it.
    # paleo/chron - idx - table_type - idx
    # Not sure what to do with m.group(2) yet
    sheets.append({
        "paleo_chron": pc,
        "pc_idx": int(m.group(2)),
        "model_idx": model_idx,
        "table_type": m.group(4),
        "table_idx": int(m.group(5)),
        "name": old_sheet,
        "filename": "{}.{}{}{}{}.csv".format(name, pc, m.group(2), m.group(4), m.group(5)),
        "table_name": "{}{}{}{}".format(pc, m.group(2), m.group(4), m.group(5)),
        "data": ""
    })

    return sheets, paleo_ct, chron_ct


def _place_tables(metadata):
    """
    All the data has been parsed, now put the data into a LiPD 1.2 structure.
    :param dict metadata:
    :return:
    """
    logger_excel.info("enter place_tables")
    paleo, chron = _create_metadata_skeleton(metadata)

    for item in metadata:
        pc_idx = item["pc_idx"] - 1
        m_idx = item["model_idx"]
        table_idx = item["table_idx"] - 1
        pc = item["paleo_chron"]
        name = item["name"]
        table_type = item["table_type"]
        data = item["data"]
        logger_excel.info("placing table: {}".format(name))

        # Can only decrement model index if it exists. Check first.
        if m_idx:
            m_idx -= 1

        if pc == "paleo":
            # If it's measurement table, it goes in first.
            if table_type == "measurement":
                paleo[pc_idx]["paleoMeasurementTable"][table_idx] = data
            # Other types of tables go one step below
            elif table_type in ("ensemble", "distribution", "model"):
                if table_type == "model":
                    paleo[pc_idx]["paleoModel"][m_idx]["paleoModelTable"] = data
                elif table_type == "ensemble":
                    paleo[pc_idx]["paleoModel"][m_idx]["ensembleTable"] = data
                elif table_type == "distribution":
                    paleo[pc_idx]["paleoModel"][m_idx]["distributionTable"][table_idx] = data

        elif pc == "chron":
            # If it's measurement table, it goes in first.
            if table_type == "measurement":
                chron[pc_idx]["chronMeasurementTable"][table_idx] = data
            # Other types of tables go one step below
            elif table_type in ("ensemble", "distribution", "model"):
                if table_type == "model":
                    paleo[pc_idx]["chronModel"][m_idx]["chronModelTable"] = data
                elif table_type == "ensemble":
                    paleo[pc_idx]["chronModel"][m_idx]["ensembleTable"] = data
                elif table_type == "distribution":
                    paleo[pc_idx]["chronModel"][m_idx]["distributionTable"][table_idx] = data

    return paleo, chron


def _create_metadata_skeleton(metadata):
    """
    Find out what the highest number index table is, and create a list of that length.
    :return list: Blank list of N indices
    """
    logger_excel.info("enter create_metadata_skeleton")

    pd_tmp = {"paleoMeasurementTable": [], "paleoModel": []}
    cd_tmp = {"chronMeasurementTable": [], "chronModel": []}
    cm_tmp = {"chronModelTable": {}, "ensembleTable": {}, "distributionTable": []}
    pm_tmp = {"paleoModelTable": {}, "ensembleTable": {}, "distributionTable": []}

    chron_numbers = {}
    paleo_numbers = {}

    # First, get the max number index for each list we need to build.
    for item in metadata:
        pc = item["paleo_chron"]
        tt = item["table_type"]
        pc_idx = item["pc_idx"]
        t_idx = item["table_idx"]
        m_idx = item["model_idx"]

        # Tree for chron types
        if pc == "chron":
            # Create an index for this table if there is not one already.
            if pc_idx not in chron_numbers:
                chron_numbers[pc_idx] = {"meas": 0, "models": 0, "model": {}}
            # Otherwise, start comparing indices to get the highest number for this table

            # If the models list idx is higher, then swap it.
            if m_idx:
                if m_idx > chron_numbers[pc_idx]["models"]:
                    chron_numbers[pc_idx]["models"] = m_idx

                # If this model index is not tracked yet, create it.
                if m_idx not in chron_numbers[pc_idx]["model"]:
                    chron_numbers[pc_idx]["model"] = {m_idx: {"ens": 0, "mod": 0, "dist": 0}}

            # Find what kind of table it is, and then increment count if it's higher than current.
            if tt == "measurement":
                if t_idx > chron_numbers[pc_idx]["meas"]:
                    chron_numbers[pc_idx]["meas"] = t_idx
            elif tt == "dist":
                if t_idx > chron_numbers[pc_idx]["model"][m_idx]["dist"]:
                    chron_numbers[pc_idx]["model"][m_idx]["dist"] = t_idx

        # Tree for paleo types
        elif pc == "paleo":
            # Create an index for this table if there is not one already.
            if pc_idx not in paleo_numbers:
                paleo_numbers[pc_idx] = {"meas": 0, "models": 0, "model": {}}
            # Otherwise, start comparing indices to get the highest number for this table

            # If the models list idx is higher, then swap it.
            if m_idx:
                if m_idx > paleo_numbers[pc_idx]["models"]:
                    paleo_numbers[pc_idx]["models"] = m_idx

                # If this model index is not tracked yet, create it.
                if m_idx not in paleo_numbers[pc_idx]["model"]:
                    paleo_numbers[pc_idx]["model"] = {m_idx: {"ens": 0, "mod": 0, "dist": 0}}

            # Find what kind of table it is, and then increment count if it's higher than current.
            if tt == "measurement":
                if t_idx > paleo_numbers[pc_idx]["meas"]:
                    paleo_numbers[pc_idx]["meas"] = t_idx
            elif tt == "dist":
                if t_idx > paleo_numbers[pc_idx]["model"][m_idx]["dist"]:
                    paleo_numbers[pc_idx]["model"][m_idx]["dist"] = t_idx

            # elif tt == "model":
            #     if t_idx > paleo_numbers[pc_idx]["model"]:
            #         paleo_numbers[pc_idx]["model"] = t_idx

            # elif tt == "ens":
            #     if t_idx > paleo_numbers[pc_idx]["model"][m_idx]["ens"]:
            #         paleo_numbers[pc_idx]["ens"] = t_idx

    # Create top lists with N tables
    paleo = [copy.deepcopy(pd_tmp)] * len(paleo_numbers)
    chron = [copy.deepcopy(cd_tmp)] * len(chron_numbers)

    for idx1, table in paleo_numbers.items():
        paleo[idx1-1]["paleoMeasurementTable"] = [None] * paleo_numbers[idx1]["meas"]
        paleo[idx1-1]["paleoModel"] = [copy.deepcopy(cm_tmp)] * paleo_numbers[idx1]["models"]
        for idx2, nums in table["model"].items():
            paleo[idx1-1]["paleoModel"][idx2-1] = {"paleoModelTable": {}, "ensembleTable": {},
                                               "distributionTable": [None] * nums["dist"]}

    for idx1, table in chron_numbers.items():
        chron[idx1-1]["chronMeasurementTable"] = [None] * chron_numbers[idx1]["meas"]
        chron[idx1-1]["chronModel"] = [copy.deepcopy(cm_tmp)] * chron_numbers[idx1]["models"]
        for idx2, nums in table["model"].items():
            chron[idx1-1]["chronModel"][idx2] = {"chronModelTable": {}, "ensembleTable": {},
                                               "distributionTable": [None] * nums["dist"]}

    return paleo, chron


# PARSE


def _parse_sheet(name, workbook, sheet):
    """
    The universal spreadsheet parser. Parse chron or paleo tables of type ensemble/model/summary.
    :param str name: Filename
    :param obj workbook: Excel Workbook
    :param dict sheet: Sheet path and naming info
    :return dict dict: Table metadata and numeric data
    """
    logger_excel.info("enter parse_sheet: {}".format(sheet["name"]))

    # Open the sheet from the workbook
    temp_sheet = workbook.sheet_by_name(sheet["name"])
    filename = sheet["filename"]

    # Store table metadata and numeric data separately
    table_name = "{}DataTableName".format(sheet["paleo_chron"])

    # Organize our root table data
    table_metadata = OrderedDict()
    table_metadata[table_name] = sheet["table_name"]
    table_metadata['filename'] = filename
    table_metadata['missingValue'] = 'NaN'

    # Store all CSV in here by rows
    table_data = {filename: []}

    # Master list of all column metadata
    column_metadata = []

    # Index tracks which cells are being parsed
    col_num = 0
    row_num = 0
    nrows = temp_sheet.nrows
    col_total = 0

    # Tracks which "number" each metadata column is assigned
    col_add_ct = 1

    header_keys = []
    variable_keys = []
    variable_keys_lower = []
    mv = ""

    # Markers to track where we are on the sheet
    var_header_done = False
    metadata_on = False
    metadata_done = False
    data_on = False
    notes = False


    try:
        # Loop for every row in the sheet
        for i in range(0, nrows):
            # Hold the contents of the current cell
            cell = temp_sheet.cell_value(row_num, col_num)

            # Skip all template lines
            if isinstance(cell, str):
                # Note and missing value entries are rogue. They are not close to the other data entries.
                if cell.lower().strip() not in EXCEL_TEMPLATE:

                    if "notes" in cell.lower() and not metadata_on:
                        # Store at the root table level
                        nt = temp_sheet.cell_value(row_num, 1)
                        if nt not in EXCEL_TEMPLATE:
                            table_metadata["notes"] = nt

                    elif cell.lower().strip() in ALTS_MV:
                        # Store at the root table level and in our function
                        mv = temp_sheet.cell_value(row_num, 1)
                        # Add if not placeholder value
                        if mv not in EXCEL_TEMPLATE:
                            table_metadata["missingValue"] = mv

                    # Variable template header row
                    elif cell.lower() in EXCEL_HEADER and not metadata_on and not data_on:

                        # Grab the header line
                        row = temp_sheet.row(row_num)
                        header_keys = _get_header_keys(row)

                        # Turn on the marker
                        var_header_done = True

                    # Start parsing data section (bottom)
                    elif data_on:

                        # Parse the row, clean, and add to table_data
                        table_data = _parse_sheet_data_row(temp_sheet, row_num, col_total, table_data, filename, mv)

                    # Start parsing the metadata section. (top)
                    elif metadata_on:

                        # Reached an empty cell while parsing metadata. Mark the end of the section.
                        if cell in EMPTY:
                            metadata_on = False
                            metadata_done = True

                            # Create a list of all the variable names found
                            for entry in column_metadata:
                                try:
                                    variable_keys.append(entry["variableName"].strip())
                                    variable_keys_lower.append(entry["variableName"].lower().strip())
                                except KeyError:
                                    # missing a variableName key
                                    pass

                        # Not at the end of the section yet. Parse the metadata
                        else:
                            # Get the row data
                            row = temp_sheet.row(row_num)

                            # Get column metadata
                            col_tmp = _compile_column_metadata(row, header_keys, col_add_ct)

                            # Append to master list
                            column_metadata.append(col_tmp)
                            col_add_ct += 1

                    # Variable metadata, if variable header exists
                    elif var_header_done and not metadata_done:

                        # Start piecing column metadata together with their respective variable keys
                        metadata_on = True

                        # Get the row data
                        row = temp_sheet.row(row_num)

                        # Get column metadata
                        col_tmp = _compile_column_metadata(row, header_keys, col_add_ct)

                        # Append to master list
                        column_metadata.append(col_tmp)
                        col_add_ct += 1

                    # Variable metadata, if variable header does not exist
                    elif not var_header_done and not metadata_done and cell:
                        # LiPD Version 1.1 and earlier: Chronology sheets don't have variable headers
                        # We could blindly parse, but without a header row_num we wouldn't know where
                        # to save the metadata
                        # Play it safe and assume data for first column only: variable name
                        metadata_on = True

                        # Get the row data
                        row = temp_sheet.row(row_num)

                        # Get column metadata
                        col_tmp = _compile_column_metadata(row, header_keys, col_add_ct)

                        # Append to master list
                        column_metadata.append(col_tmp)
                        col_add_ct += 1

                    # Numeric Data. Column metadata exists and metadata_done marker is on.
                    else:
                        try:
                            if metadata_done and cell.lower().strip() in variable_keys_lower:
                                data_on = True
                                col_total = len(column_metadata)
                        except AttributeError:
                            pass
                            # cell is not a string, and lower() was not a valid call.

            # If this is a numeric cell, 99% chance it's parsing the data columns.
            elif isinstance(cell, float) or isinstance(cell, int):
                if data_on or metadata_done:

                    # Parse the row, clean, and add to table_data
                    table_data = _parse_sheet_data_row(temp_sheet, row_num, col_total, table_data, filename, mv)

            # Move on to the next row
            row_num += 1
        table_metadata["columns"] = column_metadata
    except IndexError as e:
        logger_excel.debug("parse_sheet: IndexError: sheet: {}, row_num: {}, col_num: {}, {}".format(sheet, row_num, col_num, e))

    # If there isn't any data in this sheet, and nothing was parsed, don't let this
    # move forward to final output.
    if not table_data[filename]:
        table_data = None
        table_metadata = None

    logger_excel.info("exit parse_sheet: {}".format(sheet))
    return table_metadata, table_data


def _parse_sheet_data_row(temp_sheet, row_num, col_total, table_data, filename, mv):
    """
    Parse a row from the data section of the sheet. Add the cleaned row data to the overall table data.
    :param obj temp_sheet: Excel sheet
    :param int row_num: Current sheet row
    :param int col_total: Number of column variables in this sheet
    :param dict table_data: Running record of table data
    :param str filename: Filename for this table
    :param str mv: Missing value
    :return dict: Table data with appended row
    """
    # Get row of data
    row = temp_sheet.row(row_num)

    # In case our row holds more cells than the amount of columns we have, slice the row
    # We don't want to have extra empty cells in our output.
    row = row[:col_total]

    # Replace missing values where necessary
    row = _replace_mvs(row, mv)

    # Append row to list we will use to write out csv file later.
    table_data[filename].append(row)

    return table_data


def _replace_mvs(row, mv):
    """
    Replace Missing Values in the data rows where applicable
    :param list row: Row
    :return list: Modified row
    """
    for idx, v in enumerate(row):
        try:
            if v.value.lower() in EMPTY or v.value.lower() == mv:
                row[idx] = "NaN"
            else:
                row[idx] = v.value
        except AttributeError:
            if v.value == mv:
                row[idx] = "NaN"
            else:
                row[idx] = v.value

    return row


def _get_header_keys(row):
    """
    Get the variable header keys from this special row
    :return list: Header keys
    """
    # Swap out NOAA keys for LiPD keys
    for idx, key in enumerate(row):
        if key.value.lower() in EXCEL_KEYS:
            row[idx] = EXCEL_KEYS[key.value.lower()]
        else:
            try:
                row[idx] = key.value.lower()
            except AttributeError:
                pass

    header_keys = _rm_cells_reverse(row)
    return header_keys


def _compile_column_metadata(row, keys, number):
    """
    Compile column metadata from one excel row ("9 part data")
    :param list row: Row of cells
    :param list keys: Variable header keys
    :return dict: Column metadata
    """
    # Store the variable keys by index in a dictionary
    column = {}

    # Use the header keys to place the column data in the dictionary
    if keys:
        for idx, key in enumerate(keys):
            column[key] = row[idx].value
        column["number"] = number

    # If there are not keys, that means it's a header-less metadata section.
    else:
        # Assume we only have one cell, because we have no keys to know what data is here.
        column["variableName"] = row[0].value
        column["number"] = number

    # Add this column to the overall metadata
    column = {k: v for k, v in column.items() if v}
    return column


def _rm_cells_reverse(l):
    """
    Remove the cells that are empty or template in reverse order. Stop when you hit data.
    :param list l: One row from the spreadsheet
    :return list: Modified row
    """
    rm = []
    # Iter the list in reverse, and get rid of empty and template cells
    for idx, key in reversed(list(enumerate(l))):
        if key.lower() in EXCEL_TEMPLATE:
            rm.append(idx)
        elif key in EMPTY:
            rm.append(idx)
        else:
            break

    for idx in rm:
        l.pop(idx)
    return l


# CSV


def _write_data_csv(csv_data):
    """
    CSV data has been parsed by this point, so take it and write it file by file.
    :return:
    """
    logger_excel.info("enter write_data_csv")
    # Loop for each file and data that is stored
    for file in csv_data:
        for filename, data in file.items():

            # Make sure we're working with the right data types before trying to open and write a file
            if isinstance(filename, str) and isinstance(data, list):
                try:
                    with open(filename, 'w+') as f:
                        w = csv.writer(f)
                        for line in data:
                            w.writerow(line)
                except Exception:
                    logger_excel.debug("write_data_csv: Unable to open/write file: {}".format(filename))

    logger_excel.info("exit write_data_csv")
    return


# GEO DATA METHODS

def geometry_linestring(lat, lon, elev):
    """
    GeoJSON Linestring. Latitude and Longitude have 2 values each.
    :param list lat: Latitude values
    :param list lon:  Longitude values
    :return dict:
    """
    logger_excel.info("enter geometry_linestring")
    d = OrderedDict()
    coordinates = []
    temp = ["", ""]

    # Point type, Matching pairs.
    if lat[0] == lat[1] and lon[0] == lon[1]:
        logger_excel.info("matching geo coordinate")
        lat.pop()
        lon.pop()
        d = geometry_point(lat, lon, elev)

    else:
        # Creates coordinates list
        logger_excel.info("unique geo coordinates")
        for i in lat:
            temp[0] = i
            for j in lon:
                temp[1] = j
                coordinates.append(copy.copy(temp))
        if elev:
            for i in coordinates:
                i.append(elev)
        # Create geometry block
        d['type'] = 'Linestring'
        d['coordinates'] = coordinates
    logger_excel.info("exit geometry_linestring")
    return d


def geometry_range(crd_range, elev, crd_type):
    """
    Range of coordinates. (e.g. 2 latitude coordinates, and 0 longitude coordinates)
    :param crd_range: Latitude or Longitude values
    :param elev: Elevation value
    :param crd_type: Coordinate type, lat or lon
    :return dict:
    """

    d = OrderedDict()
    coordinates = [[] for i in range(len(crd_range))]

    # latitude
    if crd_type == "lat":
        for idx, i in enumerate(crd_range):
            coordinates[idx] = [crd_range[idx], "NaN"]
            if elev:
                coordinates[idx].append(elev)

    # longitude
    elif crd_type == "lon":
        for idx, i in enumerate(crd_range):
            coordinates[idx] = ["NaN", crd_range[idx]]
            if elev:
                coordinates[idx].append(elev)

    d["type"] = "Range"
    d["coordinates"] = coordinates

    return d


def geometry_point(lat, lon, elev):
    """
    GeoJSON point. Latitude and Longitude only have one value each
    :param list lat: Latitude values
    :param list lon: Longitude values
    :param list elev: Elevation value
    :return dict:
    """
    logger_excel.info("enter geometry_point")
    coordinates = []
    point_dict = OrderedDict()
    for idx, val in enumerate(lat):
        try:
            coordinates.append(lat[idx])
            coordinates.append(lon[idx])
        except IndexError as e:
            print("Error: Invalid geo coordinates")
            logger_excel.debug("geometry_point: IndexError: lat: {}, lon: {}, {}".format(lat, lon, e))

    coordinates.append(elev)
    point_dict['type'] = 'Point'
    point_dict['coordinates'] = coordinates
    logger_excel.info("exit geometry_point")
    return point_dict


def compile_geometry(lat, lon, elev):
    """
    Take in lists of lat and lon coordinates, and determine what geometry to create
    :param list lat: Latitude values
    :param list lon: Longitude values
    :param float elev: Elevation value
    :return dict:
    """
    logger_excel.info("enter compile_geometry")
    lat = _remove_geo_placeholders(lat)
    lon = _remove_geo_placeholders(lon)

    # 4 coordinate values
    if len(lat) == 2 and len(lon) == 2:
        logger_excel.info("found 4 coordinates")
        geo_dict = geometry_linestring(lat, lon, elev)

        # # 4 coordinate values
        # if (lat[0] != lat[1]) and (lon[0] != lon[1]):
        #     geo_dict = geometry_polygon(lat, lon)
        # # 3 unique coordinates
        # else:
        #     geo_dict = geometry_multipoint(lat, lon)
        #

    # 2 coordinate values
    elif len(lat) == 1 and len(lon) == 1:
        logger_excel.info("found 2 coordinates")
        geo_dict = geometry_point(lat, lon, elev)

    # coordinate range. one value given but not the other.
    elif (None in lon and None not in lat) or (len(lat) > 0 and len(lon) == 0):
        geo_dict = geometry_range(lat, elev, "lat")

    elif (None in lat and None not in lon) or (len(lon) > 0 and len(lat) == 0):
        geo_dict = geometry_range(lat, elev, "lon")

    # Too many points, or no points
    else:
        geo_dict = {}
        logger_excel.warn("compile_geometry: invalid coordinates: lat: {}, lon: {}".format(lat, lon))
    logger_excel.info("exit compile_geometry")
    return geo_dict


def compile_geo(d):
    """
    Compile top-level Geography dictionary.
    :param d:
    :return:
    """
    logger_excel.info("enter compile_geo")
    d2 = OrderedDict()

    # get max number of sites, or number of coordinate points given.
    num_loc = _get_num_locations(d)

    # if there's one more than one location put it in a collection
    if num_loc > 1:
        d2["type"] = "FeatureCollection"
        features = []
        for idx in range(0, num_loc):
            # Do process for one site
            site = _parse_geo_locations(d, idx)
            features.append(site)
        d2["features"] = features

    # if there's only one location
    elif num_loc == 1:
        d2 = _parse_geo_location(d)

    logger_excel.info("exit compile_geo")
    return d2


def _get_num_locations(d):
    """
    Find out how many locations are being parsed. Compare lengths of each
    coordinate list and return the max
    :param dict d: Geo metadata
    :return int: Max number of locations
    """
    lengths = []
    for key in EXCEL_GEO:
        try:
            if key != "siteName":
                lengths.append(len(d[key]))
        except Exception:
            lengths.append(1)

    try:
        num = max(lengths)
    except ValueError:
        num = 0
    return num

def _parse_geo_location(d):
    """
    Parse one geo location
    :param d:
    :return:
    """
    d2 = OrderedDict()
    filt = {}
    d2['type'] = 'Feature'
    # If the necessary keys are missing, put in placeholders so there's no KeyErrors.
    for key in EXCEL_GEO:
        if key not in d:
            d[key] = ""

    # Compile the geometry based on the info available.
    d2['geometry'] = compile_geometry([d['latMin'], d['latMax']], [d['lonMin'], d['lonMax']], d['elevation'])
    d2['properties'] = {'siteName': d['siteName']}

    return d2

def _parse_geo_locations(d, idx):
    """
    Parse one geo location
    :param d:
    :return:
    """
    d2 = OrderedDict()
    filt = {}
    d2['type'] = 'Feature'
    # If the necessary keys are missing, put in placeholders so there's no KeyErrors.
    for key in EXCEL_GEO:
        if key not in d:
            d[key] = ""

    for key in EXCEL_GEO:
        try:
            if key == "siteName" and isinstance(d["siteName"], str):
                filt["siteName"] = d["siteName"]
            else:
                filt[key] = d[key][idx]
        except KeyError:
            filt[key] = None
        except TypeError:
            filt[key] = None

    # Compile the geometry based on the info available.
    d2['geometry'] = compile_geometry([filt['latMin'], filt['latMax']], [filt['lonMin'], filt['lonMax']], filt['elevation'])
    d2['properties'] = {'siteName': filt['siteName']}

    return d2

def compile_authors(cell):
    """
    Split the string of author names into the BibJSON format.
    :param str cell: Data from author cell
    :return: (list of dicts) Author names
    """
    logger_excel.info("enter compile_authors")
    author_lst = []
    s = cell.split(';')
    for w in s:
        author_lst.append(w.lstrip())
    logger_excel.info("exit compile_authors")
    return author_lst


# MISC HELPER METHODS


def compile_temp(d, key, value):
    """
    Compiles temporary dictionaries for metadata. Adds a new entry to an existing dictionary.
    :param dict d:
    :param str key:
    :param any value:
    :return dict:
    """
    if not value:
        d[key] = None
    elif len(value) == 1:
        d[key] = value[0]
    else:
        d[key] = value
    return d


def compile_fund(workbook, sheet, row, col):
    """
    Compile funding entries. Iter both rows at the same time. Keep adding entries until both cells are empty.
    :param obj workbook:
    :param str sheet:
    :param int row:
    :param int col:
    :return list of dict: l
    """
    logger_excel.info("enter compile_fund")
    l = []
    temp_sheet = workbook.sheet_by_name(sheet)
    while col < temp_sheet.ncols:
        col += 1
        try:
            agency = temp_sheet.cell_value(row, col)
            grant = temp_sheet.cell_value(row+1, col)
            if (agency != xlrd.empty_cell and agency not in EMPTY) or (grant != xlrd.empty_cell and grant not in EMPTY):
                if agency in EMPTY:
                    l.append({'grant': grant})
                elif grant in EMPTY:
                    l.append({'agency': agency})
                else:
                    l.append({'agency': agency, 'grant': grant})
        except IndexError as e:
            logger_excel.debug("compile_fund: IndexError: sheet:{} row:{} col:{}, {}".format(sheet, row, col, e))
    logger_excel.info("exit compile_fund")
    return l


def name_to_jsonld(title_in):
    """
    Convert formal titles to camelcase json_ld text that matches our context file
    Keep a growing list of all titles that are being used in the json_ld context
    :param str title_in:
    :return str:
    """
    title_out = ''
    try:
        title_in = title_in.lower()
        title_out = EXCEL_KEYS[title_in]
    except (KeyError, AttributeError) as e:
        if "(" in title_in:
            title_in = title_in.split("(")[0].strip()
        # try to find an exact match first.
        for k, v in EXCEL_KEYS.items():
            if title_in == k:
                return v
        # if no exact match, find whatever is a closest match
        for k, v in EXCEL_KEYS.items():
            if k in title_in:
                return v
    if not title_out:
        logger_excel.debug("name_to_jsonld: No match found: {}".format(title_in))
    return title_out


def instance_str(cell):
    """
    Match data type and return string
    :param any cell:
    :return str:
    """
    if isinstance(cell, str):
        return 'str'
    elif isinstance(cell, int):
        return 'int'
    elif isinstance(cell, float):
        return 'float'
    else:
        return 'unknown'


def replace_mvs(cell_entry, missing_val):
    """
    The missing value standard is "NaN". If there are other missing values present, we need to swap them.
    :param str cell_entry: Contents of target cell
    :param str missing_val:
    :return str:
    """
    if isinstance(cell_entry, str):
        missing_val_list = ['none', 'na', '', '-', 'n/a']
        if missing_val.lower() not in missing_val_list:
            missing_val_list.append(missing_val)
        try:
            if cell_entry.lower() in missing_val_list:
                cell_entry = 'NaN'
        except (TypeError, AttributeError) as e:
            logger_excel.debug("replace_missing_vals: Type/AttrError: cell: {}, mv: {} , {}".format(cell_entry, missing_val, e))
    return cell_entry


def extract_units(string_in):
    """
    Extract units from parenthesis in a string. i.e. "elevation (meters)"
    :param str string_in:
    :return str:
    """
    start = '('
    stop = ')'
    return string_in[string_in.index(start) + 1:string_in.index(stop)]


def extract_short(string_in):
    """
    Extract the short name from a string that also has units.
    :param str string_in:
    :return str:
    """
    stop = '('
    return string_in[:string_in.index(stop)]


# DATA WORKSHEET HELPER METHODS


def cells_rt_meta_pub(workbook, sheet, row, col, pub_qty):
    """
    Publication section is special. It's possible there's more than one publication.
    :param obj workbook:
    :param str sheet:
    :param int row:
    :param int col:
    :param int pub_qty: Number of distinct publication sections in this file
    :return list: Cell data for a specific row
    """
    logger_excel.info("enter cells_rt_meta_pub")
    col_loop = 0
    cell_data = []
    temp_sheet = workbook.sheet_by_name(sheet)
    while col_loop < pub_qty:
        col += 1
        col_loop += 1
        cell_data.append(temp_sheet.cell_value(row, col))
    logger_excel.info("exit cells_rt_meta_pub")
    return cell_data


def cells_rt_meta(workbook, sheet, row, col):
    """
    Traverse all cells in a row. If you find new data in a cell, add it to the list.
    :param obj workbook:
    :param str sheet:
    :param int row:
    :param int col:
    :return list: Cell data for a specific row
    """
    logger_excel.info("enter cells_rt_meta")
    col_loop = 0
    cell_data = []
    temp_sheet = workbook.sheet_by_name(sheet)
    while col_loop < temp_sheet.ncols:
        col += 1
        col_loop += 1
        try:
            if temp_sheet.cell_value(row, col) != xlrd.empty_cell and temp_sheet.cell_value(row, col) != '':
                cell_data.append(temp_sheet.cell_value(row, col))
        except IndexError as e:
            logger_excel.warn("cells_rt_meta: IndexError: sheet: {}, row: {}, col: {}, {}".format(sheet, row, col, e))
    logger_excel.info("exit cells_right_meta")
    return cell_data


def cells_dn_meta(workbook, sheet, row, col, final_dict):
    """
    Traverse all cells in a column moving downward. Primarily created for the metadata sheet, but may use elsewhere.
    Check the cell title, and switch it to.
    :param obj workbook:
    :param str sheet:
    :param int row:
    :param int col:
    :param dict final_dict:
    :return: none
    """
    logger_excel.info("enter cells_dn_meta")
    row_loop = 0
    pub_cases = ['id', 'year', 'author', 'journal', 'issue', 'volume', 'title', 'pages',
                 'reportNumber', 'abstract', 'alternateCitation']
    geo_cases = ['latMin', 'lonMin', 'lonMax', 'latMax', 'elevation', 'siteName', 'location']

    # Temp
    pub_qty = 0
    geo_temp = {}
    general_temp = {}
    pub_temp = []
    funding_temp = []

    temp_sheet = workbook.sheet_by_name(sheet)

    # Loop until we hit the max rows in the sheet
    while row_loop < temp_sheet.nrows:
        try:
            # Get cell value
            cell = temp_sheet.cell_value(row, col)

            # If there is content in the cell...
            if cell not in EMPTY:

                # Convert title to correct format, and grab the cell data for that row
                title_formal = temp_sheet.cell_value(row, col)
                title_json = name_to_jsonld(title_formal)

                # If we don't have a title for it, then it's not information we want to grab
                if title_json:

                    # Geo
                    if title_json in geo_cases:
                        cell_data = cells_rt_meta(workbook, sheet, row, col)
                        geo_temp = compile_temp(geo_temp, title_json, cell_data)

                    # Pub
                    # Create a list of dicts. One for each pub column.
                    elif title_json in pub_cases:

                        # Authors seem to be the only consistent field we can rely on to determine number of Pubs.
                        if title_json == 'author':
                            cell_data = cells_rt_meta(workbook, sheet, row, col)
                            pub_qty = len(cell_data)
                            for i in range(pub_qty):
                                author_lst = compile_authors(cell_data[i])
                                pub_temp.append({'author': author_lst, 'pubDataUrl': 'Manually Entered'})
                        else:
                            cell_data = cells_rt_meta_pub(workbook, sheet, row, col, pub_qty)
                            for pub in range(pub_qty):
                                if title_json == 'id':
                                    pub_temp[pub]['identifier'] = [{"type": "doi", "id": cell_data[pub]}]
                                else:
                                    pub_temp[pub][title_json] = cell_data[pub]
                    # Funding
                    elif title_json == 'agency':
                        funding_temp = compile_fund(workbook, sheet, row, col)

                    # All other cases do not need fancy structuring
                    else:
                        cell_data = cells_rt_meta(workbook, sheet, row, col)
                        general_temp = compile_temp(general_temp, title_json, cell_data)

        except IndexError as e:
            logger_excel.debug("cells_dn_datasheets: IndexError: sheet: {}, row: {}, col: {}, {}".format(sheet, row, col, e))
        row += 1
        row_loop += 1

    # Compile the more complicated items
    geo = compile_geo(geo_temp)

    logger_excel.info("compile metadata dictionary")
    # Insert into final dictionary
    final_dict['@context'] = "context.jsonld"
    final_dict['pub'] = pub_temp
    final_dict['funding'] = funding_temp
    final_dict['geo'] = geo

    # Add remaining general items
    for k, v in general_temp.items():
        final_dict[k] = v
    logger_excel.info("exit cells_dn_meta")
    return


# CHRONOLOGY HELPER METHODS


def count_chron_variables(temp_sheet):
    """
    Count the number of chron variables
    :param obj temp_sheet:
    :return int: variable count
    """
    total_count = 0
    start_row = traverse_to_chron_var(temp_sheet)
    while temp_sheet.cell_value(start_row, 0) != '':
        total_count += 1
        start_row += 1
    return total_count


def get_chron_var(temp_sheet, start_row):
    """
    Capture all the vars in the chron sheet (for json-ld output)
    :param obj temp_sheet:
    :param int start_row:
    :return: (list of dict) column data
    """
    col_dict = OrderedDict()
    out_list = []
    column = 1

    while (temp_sheet.cell_value(start_row, 0) != '') and (start_row < temp_sheet.nrows):
        short_cell = temp_sheet.cell_value(start_row, 0)
        units_cell = temp_sheet.cell_value(start_row, 1)
        long_cell = temp_sheet.cell_value(start_row, 2)

        # Fill the dictionary for this column
        col_dict['number'] = column
        col_dict['variableName'] = short_cell
        col_dict['description'] = long_cell
        col_dict['units'] = units_cell
        out_list.append(col_dict.copy())
        start_row += 1
        column += 1

    return out_list


def traverse_to_chron_data(temp_sheet):
    """
    Traverse down to the first row that has chron data
    :param obj temp_sheet:
    :return int: traverse_row
    """
    traverse_row = traverse_to_chron_var(temp_sheet)
    reference_var = temp_sheet.cell_value(traverse_row, 0)

    # Traverse past all the short_names, until you hit a blank cell (the barrier)
    while temp_sheet.cell_value(traverse_row, 0) != '':
        traverse_row += 1
    # Traverse past the empty cells until we hit the chron data area
    while temp_sheet.cell_value(traverse_row, 0) == '':
        traverse_row += 1

    # Check if there is a header row. If there is, move past it. We don't want that data
    if temp_sheet.cell_value(traverse_row, 0) == reference_var:
        traverse_row += 1
    logger_excel.info("traverse_to_chron_data: row:{}".format(traverse_row))
    return traverse_row


def traverse_to_chron_var(temp_sheet):
    """
    Traverse down to the row that has the first variable
    :param obj temp_sheet:
    :return int:
    """
    row = 0
    while row < temp_sheet.nrows - 1:
        if 'Parameter' in temp_sheet.cell_value(row, 0):
            row += 1
            break
        row += 1
    logger_excel.info("traverse_to_chron_var: row:{}".format(row))
    return row


def get_chron_data(temp_sheet, row, total_vars):
    """
    Capture all data in for a specific chron data row (for csv output)
    :param obj temp_sheet:
    :param int row:
    :param int total_vars:
    :return list: data_row
    """
    data_row = []
    missing_val_list = ['none', 'na', '', '-']
    for i in range(0, total_vars):
        cell = temp_sheet.cell_value(row, i)
        if isinstance(cell, str):
            cell = cell.lower()
        if cell in missing_val_list:
            cell = 'NaN'
        data_row.append(cell)
    return data_row


def _remove_geo_placeholders(l):
    """
    Remove placeholders from coordinate lists and sort
    :param list l: Lat or long list
    :return list: Modified list
    """
    vals = []
    for i in l:
        if isinstance(i, list):
            for k in i:
                if isinstance(k, float) or isinstance(k, int):
                    vals.append(k)
        elif isinstance(i, float) or isinstance(i, int):
            vals.append(i)
    vals.sort()
    return vals


if __name__ == '__main__':
    excel()



import os

from ..helpers.directory import list_files, create_tmp_dir
from ..helpers.jsons import read_json_from_file
from ..helpers.zips import *
from .lpd_noaa import LPD_NOAA
from .noaa_lpd import NOAA_LPD
from ..helpers.loggers import create_logger

logger_noaa = create_logger("noaa")


def noaa():
    """
    Convert between NOAA and LiPD file formats.
    :return:
    """
    logger_noaa.info("enter noaa")
    dir_root = os.getcwd()
    # Run lpd_noaa or noaa_lpd ?
    print("Which conversion?\n1. LPD to NOAA\n2. NOAA to LPD\n")
    mode = input("Option: ")
    logger_noaa.info("chose option: {}".format(mode))
    f_list = []
    ft = ''

    # .lpd to noaa
    if mode == '1':
        f_list = list_files('.lpd')
        ft = ' LiPD'
    # Find all needed files in current directory
    elif mode == '2':
        f_list = list_files('.txt')
        if 'noaa-blank.txt' in f_list:
            f_list.remove('noaa-blank.txt')
        ft = ' NOAA'
    logger_noaa.info("Found {} {} file(s)".format(str(len(f_list)), ft))
    print("Found {0} {1} file(s)".format(str(len(f_list)), ft))
    # Process each available file
    for name_ext in f_list:
        print('processing: {}'.format(name_ext))
        logger_noaa.info("processing: {}".format(name_ext))

        # File name w/o extension
        name = os.path.splitext(name_ext)[0]

        # Unzip file and get tmp directory path
        dir_tmp = create_tmp_dir()

        # Process file
        if mode == '1':
            unzipper(name_ext, dir_tmp)
            _process_lpd(name, dir_tmp, dir_root)
        elif mode == '2':
            _process_noaa(name, dir_tmp, dir_root)

        # Delete tmp folder and all contents
        os.chdir(dir_root)
        shutil.rmtree(dir_tmp)

    print("Process Complete")
    logger_noaa.info("exit noaa")
    return


def _process_noaa(name, dir_tmp, dir_root):
    """
    Convert NOAA format to LiPD format
    :param str name: Name of file, no extension
    :param str dir_tmp: Path to tmp directory
    :return None:
    """
    logger_noaa.info("enter process_noaa")
    NOAA_LPD(dir_root, dir_tmp, name).main()
    os.chdir(dir_root)
    zipper(dir_tmp, name, name + ".lpd")
    os.rename(name + ".lpd" + '.zip', name + ".lpd")
    logger_noaa.info("exit process_noaa")
    return


def _process_lpd(name, dir_tmp, dir_root):
    """
    Convert a LiPD format to NOAA format
    :param str name: Name of file, no extension
    :param str dir_tmp: Path to tmp directory
    :return None:
    """
    logger_noaa.info("enter process_lpd")
    dir_bag = os.path.join(dir_tmp, name)
    dir_data = os.path.join(dir_bag, 'data')

    # Navigate down to jld file
    # Directory Change: dir_root -> dir_data
    try:
        os.chdir(dir_data)
        # Open file and execute conversion script
        d = read_json_from_file(os.path.join(dir_data, name + '.jsonld'))
        LPD_NOAA(dir_root, name, d).main()
    except FileNotFoundError:
        logger_noaa.debug("process_lpd: FileNotFound: tmp directory not found")
        print("Error: Unable to process {}".format(name))
    logger_noaa.info("exit process_lpd")
    return




import os
import tempfile
import shutil
import ntpath
import time
import tkinter
from tkinter import filedialog

from .loggers import create_logger


logger_directory = create_logger('directory')


def filename_from_path(path):
    """
    Extract the file name from a given file path.
    :param str path: File path
    :return str: File name with extension
    """
    head, tail = ntpath.split(path)
    return head, tail or ntpath.basename(head)


def create_tmp_dir():
    """
    Creates tmp directory in OS temp space.
    :return str: Path to tmp directory
    """
    t = tempfile.mkdtemp()
    logger_directory.info("temp directory: {}".format(t))
    return t


def list_files(x):
    """
    Lists file(s) in given path of the X type.
    :param str x: File extension that we are interested in.
    :return list of str: File name(s) to be worked on
    """
    logger_directory.info("enter list_files")
    file_list = []
    for file in os.listdir():
        if file.endswith(x) and not file.startswith(("$", "~", ".")):
            file_list.append(file)
    logger_directory.info("exit list_files")
    return file_list


def dir_cleanup(dir_bag, dir_data):
    """
    Moves JSON and csv files to bag root, then deletes all the metadata bag files. We'll be creating a new bag with
    the data files, so we don't need the other text files and such.
    :param str dir_bag: Path to root of Bag
    :param str dir_data: Path to Bag /data subdirectory
    :return None:
    """
    logger_directory.info("enter dir_cleanup")
    # dir : dir_data -> dir_bag
    os.chdir(dir_bag)

    # Delete files in dir_bag
    for file in os.listdir(dir_bag):
        if file.endswith('.txt'):
            os.remove(os.path.join(dir_bag, file))
    logger_directory.info("deleted files in dir_bag")
    # Move dir_data files up to dir_bag
    for file in os.listdir(dir_data):
        shutil.move(os.path.join(dir_data, file), dir_bag)
    logger_directory.info("moved dir_data contents to dir_bag")
    # Delete empty dir_data folder
    shutil.rmtree(dir_data)
    logger_directory.info("exit dir_cleanup")
    return


def check_file_age(filename, days):
    """
    Check if the target file has an older creation date than X amount of time.
    i.e. One day: 60*60*24
    :param str filename: Target filename
    :param int days: Limit in number of days
    :return bool: True - older than X time, False - not older than X time
    """
    logger_directory.info("enter check_file_age")
    # Multiply days given by time for one day.
    t = days * 60 * 60 * 24
    now = time.time()
    specified_time = now - t
    try:
        if os.path.getctime(filename) < specified_time:
            # File found and out of date
            logger_directory.info("{} not up-to-date".format(filename))
            logger_directory.info("exiting check_file_age()")
            return True
        # File found, and not out of date
        logger_directory.info("{} and up-to-date".format(filename))
        logger_directory.info("exiting check_file_age()")
        return False
    except FileNotFoundError:
        # File not found. Need to download it.
        logger_directory.info("{} not found in {}".format(filename, os.getcwd()))
        logger_directory.info("exiting check_file_age()")
        return True


def browse_dialog_dir():
    """
    Open up a GUI browse dialog window and let to user pick a target directory.
    :return str: Target directory path
    """
    logger_directory.info("enter browse_dialog")
    root = tkinter.Tk()
    root.withdraw()
    root.update()
    path = tkinter.filedialog.askdirectory(parent=root, initialdir=os.path.expanduser('~'), title='Please select a directory')
    logger_directory.info("chosen path: {}".format(path))
    root.destroy()
    logger_directory.info("exit browse_dialog")
    return path


def get_src_or_dst(mode):
    """
    User sets the path to a LiPD source location
    :return str: Path
    """
    logger_directory.info("enter set_src_or_dst")
    _path = ''
    invalid = True
    count = 0

    if mode == "save":
        prompt = "Where would you like to save your file(s)?\n1. Current\n2. Browse\n3. Downloads\n4. Notebook\n"
    elif mode == "load":
        prompt = "Where are your file(s) stored?\n1. Current\n2. Browse\n3. Downloads\n4. Notebook\n"
    else:
        # did you forget to enter a mode? silly
        invalid = False
    while invalid:
        print(prompt)
        option = input("Option: ")
        if option == '1':
            # Current directory
            logger_directory.info("1: current")
            _path = os.getcwd()
        elif option == '2':
            # Open up the GUI browse dialog
            logger_directory.info("2: browse")
            _path = browse_dialog_dir()
            # Set the path to the local files in CLI and lipd_lib
        elif option == '3':
            # Set the path to the system downloads folder.
            logger_directory.info("3: downloads")
            _path = os.path.expanduser('~/Downloads')
        elif option == '4':
            # Set path to the Notebook folder
            logger_directory.info("4: notebook ")
            _path = os.path.expanduser('~/LiPD_Notebooks')
        else:
            # Something went wrong. Prompt again. Give a couple tries before defaulting to downloads folder
            if count == 2:
                logger_directory.warn("too many attempts")
                print("Too many failed attempts. Defaulting to Downloads Folder.")
                _path = os.path.expanduser('~/Downloads')
            else:
                count += 1
                logger_directory.warn("failed attempts: {}".format(count))
                print("Invalid option. Try again.")
        if _path:
            invalid = False
    logger_directory.info("exit set_src_or_dst")
    return _path


def browse_dialog_file():
    """
    Open up a GUI browse dialog window and let to user pick a target directory.
    :return str: Target directory path
    :return:
    """
    root = tkinter.Tk()
    root.withdraw()
    root.update()
    path = tkinter.filedialog.askopenfilename(parent=root, initialdir=os.path.expanduser('~'), title='Please select a file')
    root.destroy()
    return path



import copy
import os
import json

import demjson

from .blanks import *
from .loggers import *
from .misc import *

logger_jsons = create_logger("jsons")


# IMPORT


def read_json_from_file(filename):
    """
    Import the JSON data from target file.
    :param str filename: Target File
    :return dict: JSON data
    """
    logger_jsons.info("enter read_json_from_file")
    d = {}
    try:
        # Load and decode
        d = demjson.decode_file(filename)
        logger_jsons.info("successful read from json file")
    except FileNotFoundError:
        # Didn't find a jsonld file. Maybe it's a json file instead?
        try:
            d = demjson.decode_file(os.path.splitext(filename)[0] + '.json')
        except FileNotFoundError as e:
            # No json or jsonld file. Exit
            print("Error: jsonld file not found: {}".format(filename))
            logger_jsons.debug("read_json_from_file: FileNotFound: {}, {}".format(filename, e))
        except Exception:
            print("Error: unable to read jsonld file")

    if d:
        d = remove_empty_fields(d)
    logger_jsons.info("exit read_json_from_file")
    return d


def idx_num_to_name(d):
    """
    Switch from index-by-number to index-by-name.
    :param dict d: Metadata
    :return dict: Modified Metadata
    """
    logger_jsons.info("enter idx_num_to_name")

    # Take whatever lipd version this file is, and convert it to the most current lipd version
    d = update_lipd_version(d)

    if "paleoData" in d:
        tmp_pd = _import_data(d["paleoData"], "paleo")
        d["paleoData"] = tmp_pd

    if "chronData" in d:
        tmp_cd = _import_data(d["chronData"], "chron")
        d["chronData"] = tmp_cd

    logger_jsons.info("exit idx_num_to_name")
    return d


def _import_data(section_data, pc):
    """
    Import the section metadata and change it to index-by-name.
    :param list section_data: Section metadata
    :param str pc: Paleo or Chron
    :return dict: Modified paleoData
    """
    logger_jsons.info("enter import_data: {}".format(pc))
    d = {}
    idx = 1
    try:
        for table in section_data:
            tmp_table = {}

            # Get the table name from the first measurement table, and use that as the index name for this table
            tmp_table_name = get_variable_name_table("{}DataTableName".format(pc), table, pc)

            # Process the paleo measurement table
            if "{}MeasurementTable".format(pc) in table:
                tmp_pmt = _import_meas(table["{}MeasurementTable".format(pc)], pc)
                tmp_table["{}MeasurementTable".format(pc)] = tmp_pmt

            # Process the paleo model
            if "{}Model".format(pc) in table:
                tmp_cm = _import_model(table["{}Model".format(pc)], pc)
                tmp_table["{}Model".format(pc)] = tmp_cm

            # If we only have generic table names, and one exists already, don't overwrite. Create dynamic name
            if tmp_table_name in d:
                tmp_table_name = "{}_{}".format(tmp_table_name, idx)
                idx += 1

            # Put the final product into the output dictionary. Indexed-by-table-name
            d[tmp_table_name] = tmp_table

    except AttributeError:
        # paleoData is not a list like it should be.
        logger_jsons.info("import_data: {},  AttributeError: expected list type, given {}".format(pc, type(section_data)))
    logger_jsons.info("exit import_data: {}".format(pc))
    return d


def _import_model(models, pc):
    """
    Change the nested items of the paleoModel data. Overwrite the data in-place.
    :param list models:
    :param str pc: Paleo or Chron
    :return list:
    """
    logger_jsons.info("enter import_model".format(pc))

    try:
        for model in models:
            # Keep the original dictionary, but replace the three main entries below

            # Do a direct replacement of chronModelTable columns. No table name, no table work needed.
            if "summaryTable" in model:
                model["summaryTable"]["columns"] = _idx_col_by_name(model["summaryTable"]["columns"])
            elif "{}ModelTable".format(pc) in model:
                model["{}ModelTable".format(pc)]["columns"] = _idx_col_by_name(model["{}ModelTable".format(pc)]["columns"])

            # Do a direct replacement of ensembleTable columns. No table name, no table work needed.
            if "ensembleTable" in model:
                model["ensembleTable"]["columns"] = _idx_col_by_name(model["ensembleTable"]["columns"])

            # Iter over each distribution. Check for calibrated ages (old) just in case
            if "calibratedAges" in model:
                model["distributionTable"] = _import_dist(model, "calibratedAges")
                model.pop("calibratedAges")
            elif "distributionTable" in model:
                model["distributionTable"] = _import_dist(model, "distributionTable")

    except AttributeError:
        logger_jsons.debug("import_model_{}: AttributeError: expected list type, given {} type".format(pc, type(models)))

    logger_jsons.info("exit import_model_{}".format(pc))
    return models


def _import_meas(tables, pc):
    """
    Index the measurement table by name
    :param dict tables: Measurement table data
    :return dict:
    """
    logger_jsons.info("enter import_meas: {}".format(pc))
    table_new = {}

    for table in tables:
        # Get the table name
        name_table = get_variable_name_table("{}DataTableName".format(pc), table, pc)

        # Call idx_table_by_name
        table = _idx_table_by_name(table)

        # Enter the named table in the output dictionary
        table_new[name_table] = table
    logger_jsons.info("exit import_meas: {}".format(pc))
    return table_new


def _import_dist(model, key):
    """
    Import calibratedAges or distributionTable.
    :param model: Table metadata
    :param key: calibratedAges or distributionTable
    :return: Modified table metadata
    """
    tmp_calib = {}
    for calibration in model[key]:
        # Use "name" as table name
        name_calib = get_variable_name_table("name", calibration)
        # Call idx_table_by_name
        table_new = _idx_table_by_name(calibration)
        # Set the new table by name into the WIP tmp_calib
        tmp_calib[name_calib] = table_new
    return tmp_calib


def _idx_table_by_name(d):
    """
    Switch a table of data from indexed-by-number into indexed-by-name using table names
    :param dict d: Table data
    :return dict: new idx-by-name table
    """
    try:
        # Overwrite the columns entry with named columns
        d["columns"] = _idx_col_by_name(d["columns"])
    except AttributeError:
        print("Error: 'columns' data is not a dictionary")
        logger_jsons.info("idx_table_by_name: AttributeError: expected dictionary type, given {} type".format(type(d)))
    return d


def _idx_col_by_name(l):
    """
    Iter over columns list. Turn indexed-by-num list into an indexed-by-name dict. Keys are the variable names.
    :param list l: Columns
    :return dict: New column indexed-by-name
    """
    col_new = {}

    # Iter for each column in the list
    try:
        for col in l:
            try:
                col_name = col["variableName"]
                col_new[col_name] = col
            except KeyError:
                print("Error: missing 'variableName' in column")
                logger_jsons.info("idx_col_by_name: KeyError: missing variableName key")
    except AttributeError:
        logger_jsons.info("idx_col_by_name: AttributeError: expected list type, given {} type".format(type(l)))

    return col_new


# PREP FOR EXPORT


def split_csv_json(d):
    """
    Split JSON with CSV values into separate JSON and CSV dictionaries.
    :param dict d: JSON metadata with CSV values in paleoData columns
    :return dict, dict: JSON only metadata, CSV organized by filename->column
    """
    logger_jsons.info("enter split_csv_json")
    # First, get CSV values and organize.
    csv = get_csv_from_json(d)
    # Then remove CSV values, which gives us JSON only.
    j = remove_csv_from_json(d)
    logger_jsons.info("exit split_csv_json")
    return j, csv


def get_csv_from_json(d):
    """
    Get CSV values when mixed into json data. Pull out the CSV data and put it into a dictionary.
    :param dict d: JSON with CSV values
    :return dict: CSV values. (i.e. { CSVFilename1: { Column1: [Values], Column2: [Values] }, CSVFilename2: ... }
    """
    logger_jsons.info("enter get_csv_from_json")
    csv_data = {}

    if "paleoData" in d:
        csv_data = _get_csv_from_section(d, "paleoData", csv_data)

    if "chronData" in d:
        csv_data = _get_csv_from_section(d, "chronData", csv_data)

    logger_jsons.info("exit get_csv_from_json")
    return csv_data


def _get_csv_from_section(d, pc, csv_data):
    """
    Get csv from paleo and chron sections
    :param dict d: Metadata
    :param str pc: Paleo or chron
    :return dict: running csv data
    """
    logger_jsons.info("enter get_csv_from_section: {}".format(pc))

    for table, table_content in d[pc].items():
        # Create entry for this table/CSV file (i.e. Asia-1.measTable.PaleoData.csv)
        # Note: Each table has a respective CSV file.
        csv_data[table_content['filename']] = {}
        for column, column_content in table_content['columns'].items():
            # Set the "values" into csv dictionary in order of column "number"
            csv_data[table_content['filename']][column_content['number']] = column_content['values']

    logger_jsons.info("exit get_csv_from_section: {}".format(pc))
    return csv_data


def remove_csv_from_json(d):
    """
    Remove all CSV data 'values' entries from paleoData table in the JSON structure.
    :param dict d: JSON data - old structure
    :return dict: Metadata dictionary without CSV values
    """
    logger_jsons.info("enter remove_csv_from_json")

    # Check both sections
    if "paleoData" in d:
        d = _remove_csv_from_section(d, "paleoData")

    if "chronData" in d:
        d = _remove_csv_from_section(d, "chronData")

    logger_jsons.info("exit remove_csv_from_json")
    return d


def _remove_csv_from_section(d, pc):
    """
    Remove CSV from metadata in this section
    :param dict d: Metadata
    :param str pc: Paleo or chron
    :return dict: Modified metadata
    """
    logger_jsons.info("enter remove_csv_from_json: {}".format(pc))

    for table, table_content in d[pc].items():
        for column, column_content in table_content['columns'].items():
            try:
                # try to delete the values key entry
                del column_content['values']
            except KeyError as e:
                # if the key doesn't exist, keep going
                logger_jsons.debug("remove_csv_from_json: KeyError: {}, {}".format(pc, e))

    logger_jsons.info("exit remove_csv_from_json: {}".format(pc))
    return d


# EXPORT


def write_json_to_file(filename, json_data):
    """
    Write all JSON in python dictionary to a new json file.
    :param str filename: Target json file
    :param dict json_data: JSON data
    :return None:
    """
    logger_jsons.info("enter write_json_to_file")
    json_data = remove_empty_fields(json_data)
    # Use demjson to maintain unicode characters in output
    json_bin = demjson.encode(json_data, encoding='utf-8', compactly=False)
    # Write json to file
    try:
        open(filename, "wb").write(json_bin)
        logger_jsons.info("wrote data to json file")
    except FileNotFoundError as e:
        print("Error: Writing json to file: {}".format(filename))
        logger_jsons.debug("write_json_to_file: FileNotFound: {}, {}".format(filename, e))
    logger_jsons.info("exit write_json_to_file")
    return


def idx_name_to_num(d):
    """
    Switch from index-by-name to index-by-number.
    :param dict d: Metadata
    :return dict: Modified metadata
    """
    logger_jsons.info("enter idx_name_to_num")

    # Process the paleoData section
    if "paleoData" in d:
        d["paleoData"] = _export_data(d["paleoData"], "paleo")

    # Process the chronData section
    if "chronData" in d:
        d["chronData"] = _export_data(d["chronData"], "chron")

    logger_jsons.info("exit idx_name_to_num")
    return d


def _export_data(section_data, pc):
    """
    Switch chron data to index-by-number
    :param dict section_data: Metadata
    :return list: Metadata
    """
    logger_jsons.info("enter export_data: {}".format(pc))
    l = []

    for name, table in section_data.items():

        # Process chron models
        if "{}Model".format(pc) in table:
            table["{}Model".format(pc)] = _export_model(table["{}Model".format(pc)], pc)

        # Process the chron measurement table
        if "{}MeasurementTable".format(pc) in table:
            table["{}MeasurementTable".format(pc)] = _idx_table_by_num(table["{}MeasurementTable".format(pc)])

        # Add only the table to the output list
        l.append(table)

    logger_jsons.info("exit export_data: {}".format(pc))
    return l


def _export_model(models, pc):
    """
    Switch chron model to index-by-number
    :param list models: Metadata
    :return list: modified model
    """
    logger_jsons.info("enter export_model: {}".format(pc))
    try:
        for model in models:

            if "summaryTable" in model:
                model["summaryTable"] = _idx_table_by_num(model["summaryTable"])

            if "{}ModelTable".format(pc) in model:
                model["summaryTable"] = _idx_table_by_num(model["{}ModelTable".format(pc)])
                model.pop("{}ModelTable".format(pc))

            # Process ensemble table (special two columns)
            if "ensembleTable" in model:
                model["ensembleTable"] = _idx_table_by_num(model["ensembleTable"])

            # Process calibrated ages (nested tables)
            if "calibratedAges" in model:
                ca = []
                for name, table in model["calibratedAges"].items():
                    try:
                        # Get the modified table data
                        tmp = _idx_table_by_num(table)
                        # Append it to the growing calibrated age list of tables
                        ca.append(tmp)
                    except KeyError:
                        logger_jsons.debug("export_model: {},  KeyError: missing columns key".format(pc))

                # Insert the newly built list in-place over the dictionary
                model["distributionTable"] = ca
                model.pop("calibratedAges")
            elif "distributionTable" in model:
                dt = []
                for name, table in model["distributionTable"].items():
                    try:
                        # Get the modified table data
                        tmp = _idx_table_by_num(table)
                        # Append it to the growing calibrated age list of tables
                        dt.append(tmp)
                    except KeyError:
                        logger_jsons.debug("export_model: {}, KeyError: missing columns key".format(pc))

                # Insert the newly built list in-place over the dictionary
                model["distributionTable"] = dt

    except AttributeError:
        logger_jsons.debug("export_model: {}, AttributeError: expected list type, received {} type".format(pc, type(models)))
    logger_jsons.info("exit export_model: {}".format(pc))
    return models


# NO EXPORT DISTRIBUTION?


def _idx_table_by_num(d):
    """
    Append all the values from the dictionary to an output list. Drop the keys.
    :return:
    """
    if "columns" in d:
        try:
            # Overwrite the columns dict with a new columns list
            d["columns"] = _idx_col_by_num(d["columns"])
        except AttributeError:
            print("Error: table type is incorrect")
            logger_jsons.debug("idx_table_by_num: AttributeError: expected type dict, given type {}".format(type(d)))
    return d


def _idx_col_by_num(d):
    """
    Index columns by number instead of by name. Use "number" key in column to maintain order
    :param dict d: Column data
    :return list: modified column data
    """
    l = []
    try:
        # Create an empty list that matches the length of the column dictionary
        l = [None for i in d]

        # Loop and start placing data in the output list based on its "number" entry
        for var, data in d.items():
            try:
                # Special case for ensemble table "numbers" list
                if isinstance(data["number"], list):
                    l.append(data)
                # Place at list index based on its column number
                else:
                    l[data["number"] - 1] = data
            except KeyError:
                print("Error: column is missing a 'number' key")
                logger_jsons.debug("idx_col_by_num: KeyError: missing number key")
    except AttributeError:
        logger_jsons.debug("idx_col_by_num: AttributeError: expected dict type, given {} type".format(type(d)))
    return l




import bagit
from .loggers import create_logger

logger_bagit = create_logger('bag')


def create_bag(dir_bag):
    """
    Create a Bag out of given files.
    :param str dir_bag: Directory that contains csv, jsonld, and changelog files.
    :return obj: Bag
    """
    logger_bagit.info("enter create_bag")
    try:
        bag = bagit.make_bag(dir_bag, {'Name': 'LiPD Project', 'Reference': 'www.lipd.net', 'DOI-Resolved': 'True'})
        logger_bagit.info("created bag")
        return bag
    except Exception as e:
        print("Error: failed to create bagit bag")
        logger_bagit.debug("failed to create bag, {}".format(e))
    return None


def open_bag(dir_bag):
    """
    Open Bag at the given path
    :param str dir_bag: Path to Bag
    :return obj: Bag
    """
    logger_bagit.info("enter open_bag")
    try:
        bag = bagit.Bag(dir_bag)
        logger_bagit.info("opened bag")
        return bag
    except Exception as e:
        print("Error: failed to open bagit bag")
        logger_bagit.debug("failed to open bag, {}".format(e))
    return None


def validate_md5(bag):
    """
    Check if Bag is valid
    :param obj bag: Bag
    :return None:
    """
    logger_bagit.info("validate_md5")
    if bag.is_valid():
        print("Valid md5")
        # for path, fixity in bag.entries.items():
        #     print("path:{}\nmd5:{}\n".format(path, fixity["md5"]))
    else:
        print("Invalid md5")
        logger_bagit.debug("invalid bag")
    return


def resolved_flag(bag):
    """
    Check DOI flag in bag.info to see if doi_resolver has been previously run
    :param obj bag: Bag
    :return bool: Flag
    """
    if 'DOI-Resolved' in bag.info:
        logger_bagit.info("bagit resolved_flag: true")
        return True
    logger_bagit.info("bagit resolved_flag: false")
    return False


def finish_bag(dir_bag):
    """
    Closing steps for creating a bag
    :param obj dir_bag:
    :return None:
    """
    logger_bagit.info("enter finish_bag")
    # Create a bag for the 3 files
    new_bag = create_bag(dir_bag)
    open_bag(dir_bag)
    new_bag.save(manifests=True)
    logger_bagit.info("exit finish_bag")
    return



from .doi_resolver import DOIResolver
from ..helpers.bag import *
from ..helpers.directory import *
from ..helpers.zips import *
from ..helpers.jsons import *
from ..helpers.loggers import create_logger

logger_doi_main = create_logger("doi_main")


def doi():
    """
    Main function that controls the script. Take in directory containing the .lpd file(s). Loop for each file.
    :return None:
    """
    logger_doi_main.info("enter doi_main")
    # Find all .lpd files in current directory
    # dir: ? -> dir_root
    dir_root = os.getcwd()
    f_list = list_files('.lpd')
    print("Found {0} {1} file(s)".format(str(len(f_list)), 'LiPD'))
    force = prompt_force()
    for name_ext in f_list:
        # .lpd name w/o extension
        name = os.path.splitext(name_ext)[0]

        # Unzip file and get tmp directory path
        dir_tmp = create_tmp_dir()
        unzipper(name_ext, dir_tmp)

        # Force DOI update?
        if not force:
            # Unbag and check resolved flag. Don't run if flag exists.
            if resolved_flag(open_bag(os.path.join(dir_tmp, name))):
                print('skipping: {}'.format(name_ext))
                logger_doi_main.info("skipping: {}".format(name_ext))
                shutil.rmtree(dir_tmp)

        # Process file if flag does not exist or force.
        else:
            print('processing: {}'.format(name_ext))
            logger_doi_main.info("processing: {}".format(name_ext))
            # dir: dir_root -> dir_tmp
            process_lpd(name, dir_tmp)
            # dir: dir_tmp -> dir_root
            os.chdir(dir_root)
            # Zip the directory containing the updated files. Created in dir_root directory
            zipper(dir_tmp, name, name_ext)
            os.rename(name_ext + '.zip', name_ext)
            # Cleanup and remove tmp directory
            shutil.rmtree(dir_tmp)
    logger_doi_main.info("exit doi_main")
    print("Process Complete")
    return


def process_lpd(name, dir_tmp):
    """
    Opens up json file, invokes doi_resolver, closes file, updates changelog, cleans directory, and makes new bag.
    :param str name: Name of current .lpd file
    :param str dir_tmp: Path to tmp directory
    :return none:
    """
    logger_doi_main.info("enter process_lpd")
    dir_root = os.getcwd()
    dir_bag = os.path.join(dir_tmp, name)
    dir_data = os.path.join(dir_bag, 'data')

    # Navigate down to jLD file
    # dir : dir_root -> dir_data
    os.chdir(dir_data)

    # Open jld file and read in the contents. Execute DOI Resolver.
    jld_data = read_json_from_file(os.path.join(dir_data, name + '.jsonld'))

    # Overwrite data with new data
    jld_data = DOIResolver(dir_root, name, jld_data).main()
    # Open the jld file and overwrite the contents with the new data.
    write_json_to_file(os.path.join(dir_data, name + '.jsonld'), jld_data)

    # Open changelog. timestamp it. Prompt user for short description of changes. Close and save
    update_changelog()

    # Delete old bag files, and move files to bag root for re-bagging
    # dir : dir_data -> dir_bag
    dir_cleanup(dir_bag, dir_data)
    finish_bag(dir_bag)

    logger_doi_main.info("exit process_lpd")
    return


def prompt_force():
    """
    Ask the user if they want to force update files that were previously resolved
    :return bool: response
    """
    logger_doi_main.info("enter prompt_force")
    force = False
    count = 0
    print("Do you want to force updates for previously resolved files? (y/n)")
    while True:
        f = input("> ")
        try:
            if count == 2:
                force = False
                break
            elif f.lower() in ('y', 'yes'):
                force = True
                break
            elif f.lower() in ('n', 'no'):
                force = False
                break
            else:
                print("invalid response")
        except AttributeError as e:
            print("invalid response")
            logger_doi_main.warn("invalid response: {}, {}".format(f, e))
        count += 1
    logger_doi_main.info("force update: {}".format(force))
    logger_doi_main.info("exit prompt_force")
    return force

if __name__ == '__main__':
    doi()
